/* $Header: divecomp.c,v 3.13 92/02/13 09:08:42 dave Exp $

Dive computer simulator and planning aid, by Dave Waller (dave@hpdstma.hp.com)

COPYRIGHT NOTICE (see additions below)

Copyright 1992 by Dave Waller

'divecomp' is distributed free of charge for the education and general
use of the SCUBA diving public at large.  Permission to copy and
distribute or modify the source is granted provided the copyright
notice and author attribution contained in the comments of the source,
as well as the copyright notice and author attribution message that is
displayed upon startup of the program are not removed or altered in
any way.

PLEASE NOTE:  If you make modifications to divecomp in any way, you
must retain the startup message containing the copyright notice and
author attribution.  Removal of this message is a violation of the
copyright, and heck, it is pretty unfair to me, the author, who has
invested hundreds of hours of personal time developing this program.

'divecomp' is offered without any warranty, express or implied.  The
author makes no warranty as to the accuracy or function of the
program.

SCUBA diving is a high risk activity that can result in injury or
death to the participants.  'divecomp' is intended only as an
educational tool for understanding decompression theory as it applies
to SCUBA diving.  It is _not_ intended to be used as a tool for
planning SCUBA dives, and the user is warned that dependence upon the
information provided by 'divecomp' for planning SCUBA dives can result
in serious injury or death.  SCUBA divers should utilize the tools
they were introduced to in SCUBA training for decompression
calculations and dive planning.  The author of 'divecomp' makes no
claim or warranty as to the accuracy of 'divecomp' for planning SCUBA
dives.

*/

/*
   Note this is an altered version of divecomp.c.  Specifically
   much altering has been done simply for compilation purposes.
   The algorithm should be intact as written by Dave, but this
   code should not be distributed, rather for those purposes
   an original archive should be found.

   Mike Zimmerman        zimmmt@aur.alcatel.com
*/

/*
   Minor changes to compile cleanly under HP-UX with HP ANSI C compiler.

   Wrolf Courtney        wrolf@concentric.net 5/21/98
*/

/* includes */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>	/* WC 5/21/98 */
#include <curses.h>
#include <ctype.h>
#include <fcntl.h>

/* macro definitions */

#define	NCOMP		24
#define	ONE_ATM_FSW	33.0
#define ONE_ATM_PSI	14.7
#define PSI_PER_BAR	14.5
#define	PN2		0.79
#define	ONE_ATM_FSW_N2	PN2 * ONE_ATM_FSW
#define	L_PER_CUFT	28.32
#define	LN_HALF		-0.6931
#define	INF 		3 * 24 * 3600	/* 3 days */
#define	DEPTH		0
#define	RATE		1
#define	NRATES		10
#define	DESCENT_RATE	1.0
#define	ON		1
#define	OFF		0
#define	LITER		L_PER_CUFT
#define	CUFT		1.0
#define	STATS		ncomp+1
#define	LGRAPH		1
#define	PSI		1.0
#define	BAR		0.0689
#define NOMARK		-2
#define ENGLISH		0
#define METRIC		1
#define NOCONTROL	-1
#define CNOTICE		20

#define MAX(a,b)	((a)>(b)?(a):(b))
#define ZBOUND(a)	((a)>0?(a):0)

/* Some obscure variable names and their meanings:

	bt	Bottom Time

	si	Surface Interval

	di	Depth Increment (for profile display)

*/

/* GLOBALS */

int     ncomp=0, control, lines=24, samples=0, dive=1,
	decomode=0, width=80, num_rates=1, msg, ret, num_records=0,
	iterative=OFF, graph_loading=OFF, plot_tissue=1,
	plot_mvalue=OFF, new_tank=TRUE;

float   bt=0.0, si=0.0, ap_inert, one_atm_pp_inert, ucf=1.0, di,
	last_depth=0.0, depth=0.0, s=0.0, percent_inert,
	loading(), rates[NRATES][NRATES], duration=0.0,
	total_time=0.0, dummy, rmvs=0.0, on_depth=5.0,
	off_depth=5.0, min_si=600.0, mark=NOMARK,
	profile_ucf=1.0;

typedef struct tank_type {
	char *label;
	float cap;
	float wp;
	float pres;
	float mix;
	int type;
	struct tank_type *next;
	} tank_type;

typedef struct record_type {
	float elapsed;
	float depth;
	float duration;
	float rate;
	float comp[NCOMP];
	struct {
		tank_type tank;
		int new_tank;
		float rmvs;
	} air;
	int decoflag;
	} record_type;

typedef struct model_type {
	float ongas;
	float offgas;
	float surface;
	float M0;
	float dM;
	} model_type;

typedef struct menu_type {
	char *label;
	struct menu_type *submenu;
	struct menu_type *next;
	} menu_type;

union {
	record_type *r;
	tank_type *t;
	menu_type *m;
	} sptr;


record_type record[8192];

model_type model[NCOMP];

tank_type

	tank_models[]={
	{ "AL80", 80.0, 3000.0, 3000.0, 0.79, ENGLISH, &tank_models[1] },
	{ "ST72", 72.0, 2475.0, 2475.0, 0.79, ENGLISH, &tank_models[2] },
	{ "ST92", 92.0, 2475.0, 2475.0, 0.79, ENGLISH, &tank_models[3] },
	{ "HP105", 105.0, 3500.0, 3500.0, 0.79, ENGLISH, &tank_models[4] },
	{ "HP80", 80.0, 3500.0, 3500.0, 0.79, ENGLISH, &tank_models[5] },
	{ "M12", 83.6, 2900.0, 2900.0, 0.79, METRIC, &tank_models[6] },
	{ "M18", 125.4, 2900.0, 2900.0, 0.79, METRIC, NULL }
},
	tank={ "AL80", 80.0, 3000.0, 3000.0, 0.79, ENGLISH, NULL },
	*last_tank_entry=&tank_models[6], *last_std_tank=&tank_models[6],
	*current_tank_model=&tank_models[0];

menu_type

	profile_menu[]={
	{ "Append", NULL, &profile_menu[1] },
	{ "Load", NULL, &profile_menu[2] },
	{ "Dump", NULL, &profile_menu[3] },
	{ "Log", NULL, &profile_menu[4] },
	{ "Top", NULL, NULL }
},
	air_menu[]={
	{ "Select", NULL, &air_menu[1] },
	{ "Reset", NULL, &air_menu[2] },
	{ "RMVS", NULL, &air_menu[3] },
	{ "Load", NULL, &air_menu[4] },
	{ "Build", NULL, &air_menu[5] },
	{ "Dump", NULL, &air_menu[6] },
	{ "Pres", NULL, &air_menu[7] },
	{ "Top", NULL, NULL }
},
	top_menu[]={
	{ "Profile", &profile_menu[0], &top_menu[2] },
	{ "Analysis", NULL, &top_menu[2] },
	{ "Air", &air_menu[0], &top_menu[3] },
	{ "Quit", NULL, NULL }
},
	tank_menu[]={
	{ "AL80", NULL, &tank_menu[1] },
	{ "ST72", NULL, &tank_menu[2] },
	{ "ST92", NULL, &tank_menu[3] },
	{ "HP105", NULL, &tank_menu[4] },
	{ "HP80", NULL, &tank_menu[5] },
	{ "M12", NULL, &tank_menu[6] },
	{ "M18", NULL, NULL }
},
	*last_tank_menu_entry=&tank_menu[4],

	type_menu[]={
	{ "English", NULL, &type_menu[1] },
	{ "Metric", NULL, NULL }
};
	
double	ceil(), floor(), pow(), exp(), log(), fabs();

char	unitstr[16], scale100FSWA[80], scale200FSWA[80], *cmd=NULL,

	revision[]="$Header: divecomp.c,v 3.13 92/02/13 09:08:42 dave Exp $",

	*helpstr[]={
	"\nUsage: divecomp [ -gposrulPih ] -m<file>","",
	"Valid options:\n","",
	"-a[rmvs]","Surface respiratory minute volume in cuft",
	"","Liters may be specified by appending \"l\" to",
	"","the specified value",
	"-g[%]","Percentage of inert gas in breathing mix",
	"-m[file]*","Specification of model file",
	"-s[per]","Use sample period [per] for iterative mode calculation",
	"-u[uspec]","Specify alternate depth unit and conversion",
	"-i","Use iterative calculations instead of direct (slower)",
	"-h","Display this list",
	"\n* Required option","",
	NULL
	},

	*scalestr[]={
	"",
	" 10| 20| 30| 40| 50| 60| 70| 80| 90|100|  % Allowed N2",
	" 20| 40| 60| 80|100|120|140|160|180|200|  % Allowed N2",
	" 30| 60| 90|120|150|180|210|240|270|300|  % Allowed N2",
	" 40| 80|120|160|200|240|280|320|360|400|  % Allowed N2",
	" 50|100|150|200|250|300|350|400|450|500|  % Allowed N2"
	};


/* END GLOBALS */

main(argc,argv)
int argc;
char *argv[];
{	
	FILE	*modelfp=NULL, *fopen(), *open_file(), *fp;

	int	i, opt, offgas_flag=FALSE, count;

	char    *getenv(), *envstr, *cptr, *strcpy(), *strchr(),
		mname[32], *strcat(), *strtok(), chr, *showtime(),
		*get_menu_choice(), buf[128], buf2[128], sbuf[128],
		buf3[128];

	float	update_tissues(), rate, tmp, t, find_rate(), d_tmp,
		d_target, t_sfc, readtime();

	double	atof(), fabs();

	void    update_record(), update_stats(), display_loading(),
		update_ndl_limits(), update_dive_profile(),
		display_menu(), analysis(), mode_select(), air_ops(),
		display_air(), clear_stats_region();

	extern char *optarg;

	record_type vals;

	/* initialize some defaults */

	strcpy(unitstr,"ft");
	ap_inert = one_atm_pp_inert = ONE_ATM_FSW_N2;
	percent_inert = PN2;

	/* find out how many lines display has */

	if ((envstr=getenv("LINES"))!=0) lines=atoi(envstr);
	if ((envstr=getenv("WIDTH"))!=0) width=atoi(envstr);

	/* process command line options */

	while ((opt=getopt(argc,argv,"a:g:m:s:u:ih"))!=EOF) {
		switch(opt) {
			case 'a':
				rmvs = atof(optarg) / 60.0;
				if (optarg[strlen(optarg)-1] == 'l') {
					rmvs /= L_PER_CUFT;
				}
				break;

			case 'g':
				percent_inert = atof(optarg) / 100.0;
				ap_inert =
					one_atm_pp_inert =
					percent_inert * ONE_ATM_FSW;
				break;

			case 'm':
				if ((modelfp=fopen(optarg,"r"))==NULL) {
					 perror(optarg);
					 exit(1);
				}
				strcpy(mname,optarg);
				break;

			case 's':
				s = atof(optarg);
				break;

			case 'u':
				cptr = strchr(optarg,(int)':');
				*cptr = '\0';
				ucf = atof(cptr+1);
				strcpy(unitstr, optarg);
				break;

			case 'i':
				iterative=ON;
				break;

			case 'h':
				for (i=0; helpstr[i]; i++)
					printf("%-15s%s\n",
					helpstr[i++], helpstr[i]);
				exit(1);
		}
	}

	if (modelfp==NULL) {
		fprintf(stderr, "You must specify a model with the \"-m\" option\n");
		exit(1);
	} 

	if (rmvs == 0.0) {
		fprintf(stderr, "You must specify a surface equivalent breathing rate with the\n\"-a\" option.\n");
		exit(1);
	}

	/* set up the pressure scales */

	for (i=1; i<11; i++) {
		sprintf(buf,"%3d|",(int)(i*10*ucf));
		strcat(scale100FSWA,buf);
		sprintf(buf,"%3d|",(int)(i*20*ucf));
		strcat(scale200FSWA,buf);
	}

	sprintf(buf,"  %s sea water abs.", unitstr);
	strcat(scale100FSWA,buf);
	strcat(scale200FSWA,buf);

	/* initialize the ascent rate model, and the dM values, in case
	they are not specified in the model file */

	for (i=0; i<num_rates; i++) {
		rates[i][DEPTH] = 600.0;
		rates[i][RATE] = -1.0;
	}

	for (i=0; i<NCOMP; i++) model[i].dM = 1.0;

	/* read in the computer model */

	ncomp = 0;
	while(getaline(modelfp, buf) != EOF) {

		cptr = strtok(buf,":");

		if (!strcmp(cptr, "ascent model")) {
			num_rates = 0;
			while ((cptr=strtok((char *)NULL,"<"))!=NULL) {
				rates[num_rates][RATE] = atof(cptr) / -60.0;
				rates[num_rates++][DEPTH] =
					atof(strtok((char *)NULL,":"));
			}
		}
		else if (!strcmp(cptr, "halftimes")) {
			for (ncomp=0;
			     (cptr=strtok((char *)NULL,":"))!=NULL;
			     ncomp++) {
				model[ncomp].ongas = atof(cptr) * 60.0;

				/* we copy the specified halftimes into
				the feilds for offgas halftimes and
				surface halftimes in case these values
				are not specified in the model */

				model[ncomp].offgas = model[ncomp].ongas;
				model[ncomp].surface = model[ncomp].ongas;
			}
		}
		else if (!strcmp(cptr, "offgas halftimes")) {
			for (i=0; (cptr=strtok((char *)NULL,":"))!=NULL; i++)
				model[i].offgas = atof(cptr) * 60.0;
		}
		else if (!strcmp(cptr, "surface halftimes")) {
			for (i=0; (cptr=strtok((char *)NULL,":"))!=NULL; i++)
				model[i].surface = atof(cptr) * 60.0;
		}
		else if (!strcmp(cptr, "M0 values")) {
			for (i=0; (cptr=strtok((char *)NULL,":"))!=NULL; i++)
				model[i].M0 = atof(cptr);
		}
		else if (!strcmp(cptr, "dM values")) {
			for (i=0; (cptr=strtok((char *)NULL,":"))!=NULL; i++)
				model[i].dM = atof(cptr);
		}
		else if (!strcmp(cptr, "on depth"))
			on_depth = atof(strtok((char *)NULL,":"));
		else if (!strcmp(cptr, "off depth"))
			off_depth = atof(strtok((char *)NULL,":"));
		else if (!strcmp(cptr, "min si"))
			min_si = atof(strtok((char *)NULL,":"));
		else if (!strcmp(cptr, "sample period"))
			s = atof(strtok((char *)NULL,":"));
		else {
			fprintf(stderr,"Bad entry in model file:\n\n\t%s\n\nignored.\n",buf);
			msg=1;
		}
	}

	/* check to see if there was a specification for offgassing
	halftimes. If not, copy the ongassing halftimes into the
	offgassing halftimes array */

	if (!offgas_flag) for (i=0; i<ncomp; i++)
		model[i].offgas = model[i].ongas;

	if (s == 0.0) s = 60.0; /* default sample period to 60 seconds
				    if no sample period was specified */

	/* if there was a bad entry in the model file, puase for 5
	seconds so the user can read the messages */

	if (msg) sleep(5L);
	fclose(modelfp);

	/* determine the depth increment used in the profile display */

	di = 10;

	/* intialize the elapsed time feild in the first entry in the
	   record. This needs to be set because the procedure
	   update_record() uses this value to calculate the n+1 value
	   on each invocation. */

	record[0].elapsed = 0.0;
	record[0].air.rmvs = rmvs;
	record[0].air.tank = tank;
	for (i=0; i<ncomp; i++) record[0].comp[i] = ONE_ATM_FSW_N2;

	/* initialize the display */

	init_display();

	for (;;) {
		/* Update the displays */

		record_query(total_time, &vals);
		vals.air.tank = tank;
		vals.air.rmvs = rmvs;
		update_ndl_limits(total_time);
		display_loading(vals.comp, depth, tank.mix);
		display_air(&vals);

		update_dive_profile(total_time - 3600.0, total_time, 60.0);

		refresh();

		cmd = get_menu_choice("Select:", top_menu);

		if (!strcmp(cmd,"Quit")) {
			endwin();
			exit(0);
		}
		else if (!strcmp(cmd,"Profile/Append")) {

			clear_stats_region();
			update_stats();

			while (get_sample(&d_target, &rate, &duration)) {
				
				/* relink the Analysis option back into
				the main menu if there is some profile
				to analyze */

				if ((d_target > 0) && (duration > 0))
					top_menu[0].next = &top_menu[1];

				/* update ambient N2 pressure */

				ap_inert = tank.mix * (d_target+ONE_ATM_FSW);

				/* update tissue (compartment) N2
				levels.  update_tissues() returns the
				actual final depth after
				ascending/descending to the target
				depth, which might not actually be the
				target depth if the specified duration
				is too short to fully ascend/descend
				to the target.  As such, the 'depth'
				variable must be updated with the
				return value of update_tissues().  */

				if (rate !=  0.0) {

					/* The user has specified an
					ascent/descent directly,
					bypassing the default ascent
					model.  */

					t_sfc = floor(100.0 * last_depth /
						-rate + 0.5) / 100.0;

					if ((t_sfc < duration) &&
						(rate < 0.0)) {

						/* this is an ascent
						that will reach the
						surface before the
						duration specified has
						expired.  As such, it
						must be done in two
						pieces, one for the
						actual ascent, and the
						balance ata constant
						depth of 0.0 */

						update_record(last_depth, rate,
							t_sfc, vals.comp);
						adjust_loading(vals.comp, model,
							last_depth, rate,
							t_sfc, &tank);
						update_record(0.0, 0.0,
							duration - t_sfc,
							vals.comp);
						adjust_loading(vals.comp,
							model, 0.0,
							0.0, duration -
							t_sfc, &tank);

						/* adjust the current
						depth value to reflect
						the ascent/descent */

						depth = 0.0;
					}
					else {

						/* the ascent/descent
						will not reach the
						surface, so calculate
						normally */

						update_record(last_depth, rate, duration, vals.comp);
						adjust_loading(vals.comp,
							model, last_depth,
							rate, duration, &tank);

						/* adjust the current
						depth value to reflect
						the ascent/descent */

						depth = floor(100.0 *
						(last_depth + rate * duration)
						+ 0.5) / 100.0;
					}

				}

				/* Otherwise, rate == 0.0 so we need
				to perform the normal ascent/descent
				as specified by the model */

				else {
					for (d_tmp = last_depth, t = duration;
					     (d_tmp != d_target) && (t > 0.0);) {

						/* find the applicable
						ascent/descent rate
						based upon the current
						value of depth (d_tmp)
						*/

						rate = find_rate(d_tmp, d_target, &tmp);
						tmp = (t<tmp?t:tmp);
						update_record(d_tmp, rate,
							tmp, vals.comp);
						adjust_loading(vals.comp,
							model, d_tmp,
							rate, tmp, &tank);
						t -= tmp;
						d_tmp += rate * tmp;
						d_tmp = (d_tmp<fabs(rate) ? d_target:d_tmp);
					}
					depth = d_tmp; /* update 'depth' with the actual */
						       /* depth reached */

					/* after ascent/descent has
					been accounted for, there may
					still be some of the specified
					interval left over to be spent
					at constant depth (in fact, it
					is quite likely).  Therefore,
					we need to again update the
					tissues according to the
					remaining interval.  */

					if (t > 0.0) {
						update_record(depth, 0.0,
							t, vals.comp);
						adjust_loading(vals.comp,
							model, depth, 0.0,
							t, &tank);
					}
				}

				/* update time values */

				if (depth > 6.0) {

					/* all depths >6ft are counted as bottom time */

					if ((si > 0.0) && (si < 300.0)) {

						/* We must account for
						brief periods of
						surfacing during the
						dive.  The surface
						interval counter
						starts immediately
						upon ascending to
						depths shallower than
						6 feet, yet if the
						diver returns to
						depths deeper than 6
						feet within 5 minutes,
						the time elapsed as
						surface interval must
						be added into the
						bottom time value, and
						the surface interval
						reset to 0 */

						bt += si + duration;
						si = 0.0;
					}
					else if (si >= 300.0) {	  

						/* else the diver has
						   been on the surface
						   for more than 5
						   minutes, so we
						   treat this as a new
						   dive */

						dive++;
						bt = duration;
						si=0.0;
					}

					/* otherwise, this interval is
					normal bottom time, and should
					be added to the total bottom
					time */

					else bt += duration;
				}

				/* Otherwise, the interval is at less
				than 6 feet, and should be counted
				toward surface interval */

				else si += duration;

				duration=0.0;

				last_depth = depth;
				record_query(total_time, &vals);
				update_ndl_limits(total_time);
				display_loading(vals.comp, depth, tank.mix);

				update_stats();

				update_dive_profile(total_time - 3600.0, total_time, 60.0);

				refresh();
			}
		}
		else if (!strcmp(cmd, "Profile/Dump")) {

			fp = open_file("Output filename: ", "w");

			if (fp) {
				fprintf(fp,"unit conversion:ft/%s:%.4f\n",
					unitstr, ucf);
				for (i=0, count=0; i<num_records; i++)
					if (floor(record[i].duration) > 0.0) {
						fprintf(fp,
						 "%5.1f%9s%7.1f%10s%10.5f",
						 record[i].depth * ucf,
						 showtime(record[i].duration),
						 record[i].rate * 60.0 * ucf,
						 record[i].air.tank.label,
						 record[i].air.rmvs*60.0);
					
						if (record[i].air.new_tank)
							fprintf(fp, "%8.1f", record[i].air.tank.pres);
						fprintf(fp, "\n");
						count++;
					}
				move(STATS+2,15);
				clrtoeol();
				sprintf(buf, "Total records: %d", count);
				addstr(buf);
				refresh();
				sleep(1L);
			}
			fclose(fp);
		}
		else if (!strcmp(cmd, "Profile/Load")) {

			fp = open_file("Profile filename: ", "r");

			if (fp) {
				getc(fp);
				getaline(fp, buf);
				if ((cptr=strtok(buf,":"))!=NULL) {
					if (!strcmp(cptr,
						"unit conversion")) {
						strtok(NULL, ":");
						profile_ucf =
						atof(strtok(NULL, ":"));
					}
				}
				record_query(total_time, &vals);
				while (getaline(fp, buf)!=EOF) {
					depth = atof(strtok(buf," ")) /
						profile_ucf;
					strcpy(buf2,strtok(NULL," ")); 
					rate = atof(strtok(NULL," ")) / 60;
					strcpy(buf3,strtok(NULL," "));
					rmvs = atof(strtok(NULL," ")) / 60;
					cptr = strtok(NULL, " ");
					if (strcmp(tank.label, buf3)) {
						do {
							sptr.t = tank_models;
							while ((sptr.t != NULL) && strcmp((sptr.t)->label, buf3)) sptr.t = (sptr.t)->next;
							if (sptr.t == NULL) {
								move(STATS+2, 15);
								clrtoeol();
								sprintf(sbuf, "tank model %s unknown -- Build or load a model:", buf3, buf3);
								air_ops(cmd=get_menu_choice(sbuf, air_menu));
								free(cmd);
							}
						} while (sptr.t == NULL);
						tank = *(sptr.t);
					}
					duration = readtime(buf2);
					if (cptr != NULL)
						tank.pres = atof(cptr);
					update_record(depth, rate, duration,
						vals.comp);
					adjust_loading(vals.comp, model,
						depth, rate, duration, &tank);
				}
				record_query(total_time, &vals);
				last_depth = depth = vals.depth;
				fclose(fp);

				/* link the analysis selection into the
				main menu, as there is some profile to
				analyze now */

				top_menu[0].next = &top_menu[1];
			}
		}
		else if (!strcmp(cmd, "Analysis")) analysis();

		else if (!strcmp(strtok(cmd, "/"), "Air"))
			air_ops(strtok(NULL, "/"));
		/* free memory used to return the last menu selection
		string */

		free(cmd);
	}
}

void update_record(d, r, t, comp)
float d, r, t, comp[];
{
	int i;

	float calc_air();

	for (i=0; i<ncomp; i++) record[num_records].comp[i] = comp[i];
	record[num_records].depth = d;
	record[num_records].rate = r;
	record[num_records+1].rate = r;
	record[num_records].duration = t;
	record[num_records].air.rmvs = rmvs;
	record[num_records].air.new_tank = new_tank;
	record[num_records].air.tank = tank;
	if ((d == 0.0) && (r == 0.0)) record[num_records].air.tank.mix = PN2;

	/* the next set of parameters are calculated based upon values
	from the previous record, so we need to special-case the very
	first record (record[0]). It's initial parameters were set up in
	main() while initializing others variables */

	if (num_records > 0) record[num_records].elapsed =
		record[num_records-1].elapsed + record[num_records-1].duration;
	else record[0].elapsed = 0.0;

	/* update current tank pressure */

	if ((d > 0.0) || (r != 0.0)) tank.pres -= calc_air(d, r, t, rmvs) /
		tank.cap * tank.wp;

	total_time = record[num_records].elapsed +
		record[num_records].duration;
	num_records++;
	new_tank = FALSE;
}

float update_tissues(comp, d0, d_target, t)
float comp[], d0, d_target, t;
{
	float tmp, rate, find_rate();

	double fabs();

	/* first handle the degenerate case of t=0 */

	if (t == 0.0) return(d0);

	/* this is an ascent. We must do the loading adjustments
	piecewise over the entire ascent interval, as the ascent
	model may have a variable, depth dependent ascent rate.
	*/

	while ((d_target != d0) && (t > 0.0)) {

		/* find the applicable ascent/descent rate
		based upon the current value of depth (d0) */

		rate = find_rate(d0, d_target, &tmp);
		tmp = (t<tmp?t:tmp);
		adjust_loading(comp, model, d0, rate, tmp, &tank);
		t -= tmp;
		d0 += rate * tmp;
		d0 = (d0 < fabs(rate) ? d_target : d0);
	}

	/* after ascent/descent has been accounted for, there
	may still be some of the specified interval left over
	to be spent at constant depth (in fact, it is quite
	likely). Therefore, we need to again update the tissues
	according to the remaining interval. */

	if (t > 0.0) adjust_loading(comp, model, d0, 0.0, t, &tank);

	/* return the actual depth reached. This may not be the target
	depth if the specified interval was too short to make it
	according to the ascent/descent parameters of the model */

	return(d0);
}

float find_rate(d1, d2, t)
float d1, d2, *t;
{
	int i;

	if (d2 == d1) {
		*t = 0.0;
		return(0.0);
	}

	if (d2 > d1) {

		/* this is a descent. Calculate descent portion at
		default descent rate */

		*t = (d2 - d1) / DESCENT_RATE;
		return(DESCENT_RATE);
	}
	else {

		/* find the applicable ascent rate based upon the
		current value of depth (d1) */

		for(i=0; i<num_rates-1; i++)
			if (d1<=rates[i][DEPTH]) break;

		if (i>0) {
			*t = ((d2>rates[i-1][DEPTH] ?
				d2:rates[i-1][DEPTH]) -
				d1) / rates[i][RATE];
			return(rates[i][RATE]);
		}
		else {
			*t = (d2 - d1) / rates[0][RATE];
			return(rates[0][RATE]);
		}
	}
}

adjust_loading(comp, m, d, r, t, tank)
float comp[], d, r, t;
model_type m[];
tank_type *tank;
{
	/* Adjustment of each compartment is based on the equation

	P(t) = P0 + (PA - P0 + r*Th/ln(0.5))*(1 - 0.5^(t/Th)) + r*t

	where,

	P0 = Initial pressure of compartment (fsw)
	PA = Initial ambient pressure (fsw)
	r  = ascent/descent rate
	Th = compartment halftime
	*/

	int i;

	char msg[64];

	float ambient_inert, r_inert, halftime;

	ambient_inert = tank->mix * (d + ONE_ATM_FSW);
	r_inert = tank->mix * r;

	for (i=0; i<ncomp; i++) {

		if (d == 0.0) halftime = m[i].surface;
		else if (ambient_inert > comp[i]) halftime = m[i].ongas;
		else halftime = m[i].offgas;
		comp[i] = comp[i] + (ambient_inert - comp[i] + r_inert *
			halftime / LN_HALF) * (1.0 - pow(0.5, t / halftime)) +
			r_inert * t;
	}
}

void update_stats()
{
	int limit;

	float time_to_surface(), tmp, calc_ceiling(), time_at_stop,
	      ceiling, ndl();

	char *showtime(), buf[80]; 

	record_type vals;

	void clear_stats_region(), display_air();

	clear_stats_region();

	record_query(total_time, &vals);
	vals.air.tank = tank;
	vals.air.rmvs = rmvs;

	display_air(&vals);

	sprintf(buf, "Dive: %d  Depth: %.1f %s  BT: %s ",
		dive, ucf*depth, unitstr,  showtime(bt));
	move(STATS+1,15);
	addstr(buf);

	/* find the NDL and controlling compartment for the present
	depth */

	limit = ndl(depth, vals.comp, model, &control, vals.air.tank.mix);
	if (limit > 0) {
		sprintf(buf,"NDL: %s  Sfc Int: %s",
			showtime((float)limit), showtime(si));
		move(STATS+2,15);
		addstr(buf);
	}
	else {
#ifndef SIMPLE
		attron(A_REVERSE);
#endif

		tmp = time_to_surface(depth, &time_at_stop);
		sprintf(buf,"Ceiling: %.1f %s",
			ucf*(ceiling=calc_ceiling(vals.comp, model)), unitstr);

		addstr(buf);

		move(STATS+2,15);
		sprintf(buf,
			"Next stop: %s @ %.1f %s Time to surface: %s",
			showtime(time_at_stop),
			ceil((double)ceiling/(double)5.0) * 5.0,
			unitstr,
			showtime(tmp));
		addstr(buf);

#ifndef SIMPLE
		attroff(A_REVERSE);
#endif
	}
}

void clear_stats_region()
{
	int i;

	for (i=0; i<4; i++) {
		move(STATS+i, 15);
		clrtoeol();
	}
}

/*******************************************************************

procedure name:	update_dive_profile
parameters:	t	:	Absolute time value from which to
				display the profile.

		cursor	:	position onthe screen where the cursor
				should be display

		res	:	resolution of display

return value:	none

Description:	This routine displays a 60 minute window of the dive
		profile, beginning at a specified time value from the
		beginning of the simulator run.

*********************************************************************/
void update_dive_profile(t, cursor, xres)
float t, cursor, xres;
{
	int i, j, x, y, len;
	float tmp, t1;
	float d_start, d_end;
	record_type vals;
	void depth_range(), display_depths();

	if (t<0.0) t=0.0;

	/* First erase display */

	for (i=ncomp+5; i<lines; i++) {
		move(i,10);
		clrtoeol();
	}

	/* now determine the depth of the cursor location on the profile.
	the depths along the side of the profile display will depend on
	the cursor depth. What we actually do here is call a routine
	that returns the display range for the profile; this is because
	the exact same routine is need by the ndl() routine. */

	depth_range(cursor, &d_start, &d_end);
	display_depths(d_start, d_end);

	for (	t1=t;
		t1 < (total_time - t < xres*60 ? total_time+xres : t+xres*61);
		t1 += xres) {

		x = 10 + (t1 - t) / xres;
		record_query(t1, &vals);

		if ((vals.depth >= d_start) && (vals.depth <= d_end)) {

			/* point is within the displayed area */

			y = ncomp + 5 + ceil((vals.depth - d_start) / di);
			move(y, x);
#ifndef SIMPLE
			if (t1 == floor(cursor / xres + 0.5) * xres)
				attron(A_REVERSE);
			if ((t1 == floor(mark / xres + 0.5) * xres) &&
				(mark != NOMARK)) addch('M');
			else if (vals.decoflag) addch('>');
			else addch('*');
			if (fabs(t1-cursor)<xres) attroff(A_REVERSE);
#else
			if ((t1 == floor(mark / xres + 0.5) * xres) &&
				(mark != NOMARK)) addch('M');
			else if (t1 == floor(cursor / xres + 0.5) * xres)
				addch('#');
			else if (vals.decoflag) addch('>');
			else addch('*');
#endif
			if (graph_loading) {
				if (
				   (plot_mvalue) &&
				   (model[plot_tissue-1].M0 >= d_start) &&
				   (model[plot_tissue-1].M0 <= d_end)
				) {
					y = ncomp + 5 +
						ceil(
						(model[plot_tissue-1].M0 -
						d_start) / di);
					move(y,x);
					addch('-');
				}
				if (
				   (vals.comp[plot_tissue-1] >= d_start) &&
				   (vals.comp[plot_tissue-1] <= d_end)
				) {
					y = ncomp + 5 +
						ceil((vals.comp[plot_tissue-1] -
						d_start) / di);
					move(y,x);
#ifndef SIMPLE
					if (fabs(t1-cursor)<xres)
						attron(A_REVERSE);
#endif
					if (plot_tissue < 10)
						addch(plot_tissue+'0');
					else addch(plot_tissue-10+'A');
#ifndef SIMPLE
					if (fabs(t1-cursor)<xres)
						attroff(A_REVERSE);
#endif
				}
			}
		}
	}
}

void update_ndl_limits(t)
float t;
{
	int i;

	float d, tmp, start, end;
	float ndl();	/* WC 5/21/98 */

	char buf[10];

	void depth_range();

	record_type vals;

	record_query(t, &vals);

	depth_range(t, &start, &end);

	for (d = start, i=ncomp+5; d <= end; d += di, i++) {
		tmp = ndl(d, vals.comp, model, &dummy, vals.air.tank.mix);

		move(i,0);
		if (tmp >= INF) {
			addstr("INF");
		} else if (tmp >= 60000.0) {
			addstr("***");
		} else if (tmp > 29.0) {
			sprintf(buf,"%3.0f", ceil(tmp/60.0));
			addstr(buf);
		} else addstr("   ");
	}
}

void display_loading(comp, d, mix)
float comp[], d, mix;
{
	int i, j, k, temp, scale=1;
	static int first=1;
	float loading();
	char buf[80];
	float ndl();	/* WC 5/21/98 */

	/* determine the scale to use for the display */

	for (i=0; i<ncomp; i++)
		scale = MAX(scale, (int)ceil(loading(comp[i], model[i].M0)));
	move(0,10);
	clrtoeol();
	addstr(scalestr[scale]);

	(void)ndl(d, comp, model, &control, mix);

	/* display tissue saturation */

	for (i=0; i<ncomp; i++) {

		/* display the compartment number, highlighting if it
		is the controlling tissue. */

#ifdef SIMPLE
		move(1+i,6);
		if (i==control) sprintf(buf,"*%d",i+1);
		else sprintf(buf," %d",i+1);
#else
		move(1+i,7);
		if (i==control) attron(A_REVERSE);
		sprintf(buf,"%d",i+1);
#endif
		addstr(buf);
#ifndef SIMPLE
		attroff(A_REVERSE);
#endif

		/* display the ingas/outgas indicator for the
		compartment. Ingas/outgas indicator is adjusted to 3 decimal
		place accuracy. */

		move(1+i,9);
		temp = ((d + ONE_ATM_FSW) * mix - comp[i]) * 1000.0;
		if (temp > 0) addch('>');
		if (temp == 0) addch(' ');
		if (temp < 0) addch('<');

		/* clear the bar before displaying the new value */

		move(1+i,10);
		clrtoeol();
		k = ceil(floor(1000.0 * loading(comp[i], model[i].M0)) /
			scale * 0.04);

#ifndef SIMPLE
		attron(A_REVERSE | A_UNDERLINE);
#endif
		sprintf(buf," %4.2f",ucf*comp[i]);
		for (j=0; j<k; j++) {
			if (j==(width-11-strlen(buf))) {
				addch('>');
				break;
			}	
#ifdef SIMPLE
			if (j<40/scale) addch('#');
#else
			if (j<40/scale) addch(' ');
#endif
			else addch('*');
		}
#ifndef SIMPLE
		attroff(A_REVERSE | A_UNDERLINE);
#endif
		addstr(buf);
	}
	if (first) {
		move(3, CNOTICE);
		addstr("'divecomp', decompression model");
		move(4, CNOTICE);
		addstr("simulator, by Dave Waller");
		move(5, CNOTICE);
		addstr("Copyright 1992 by Dave Waller");
		first = 0;
	}
}

init_display()
{
	int i;
	char buf[80];
	float d_start, d_end;
	void display_depths();
	void depth_range();

	initscr();

#ifndef CONVEX
	cbreak();
#else
	crmode();
#endif
	noecho();

	/* print the saturation %age scale */

	move(0,10);
	addstr(scalestr[1]);

	/* label the Dive profile graph */

	move(STATS+2,0);
	addstr("NDL  Depth");
	move(STATS+3,0);
	addstr("----------");

	/* create the dive profile display */


	depth_range(0.0, &d_start, &d_end);
	display_depths(d_start, d_end);
}

int running=OFF;
get_sample(d, r, t)
float *d,  *r, *t;
{
	float ceiling, readtime(), time_at_stop, trans_time(), calc_ceiling();
	double atof();
	char *tmp, *showtime(), *strtok(), buf[80], buf2[16], *prompt();
	int i;
	static float duration2, d_target;
	record_type vals;
	float time_to_surface();	/* WC 5/21/98 */

	record_query(total_time, &vals);

	if (iterative && running) {
		duration2 -= s;
		if (duration2 <= 0.0) {
			duration2 = 0.0;
			running = OFF;
		}
		*d = d_target;
		*r = 0.0;
		*t = s;
		return;
	}

	/* check for autodecompression mode */

	if (decomode) {
		ceiling = calc_ceiling(vals.comp, model);
		if (ceiling == 0.0) decomode = 0;
		else {
			d_target = ceil(ceiling / 5.0) * 5.0;
			time_to_surface(d_target, &time_at_stop);
			*t = ceil(time_at_stop) + 1.0;
			*d = d_target;
			*r = 0.0;
		}	
		return;
	}

	/* read depth and duration from user */

	move(STATS+3,15);
	clrtoeol();
	sprintf(buf, "Depth (default: %3.1f %s): ",
		last_depth*ucf, unitstr);
	strcpy(buf, prompt(buf));
	switch (buf[0]) {
		case 'q':
		case 'Q':
			return(0);

		case 'D':

			/* autodecompression */

			decomode = 1;
			move(STATS+3,15);
			clrtoeol();
#ifndef SIMPLE
			attron(A_REVERSE);
#endif
			addstr("*** autodecompressing ***");
#ifndef SIMPLE
			attroff(A_REVERSE);
#endif
			ceiling = calc_ceiling(vals.comp, model);
			d_target = ceil(ceiling / 5.0) * 5.0;
			time_to_surface(vals.depth, &time_at_stop);
			*t = ceil(time_at_stop) + 1.0;
			*d = d_target;
			*r = 0.0;
			return;

		case 'a':

			/* user specidifed ascent rate */

			if (strchr(buf,':') != NULL) {
				*r = atof(strtok(buf+1, ":")) / -60.0;
				*t = (atof(strtok(NULL, ":")) -
					last_depth) / *r;
				return(1);
			}
			else *r = atof(buf+1) / -60.0;
			break;

		case 'd':

			/* user specified descent rate */

			if (strchr(buf,':') != NULL) {
				*r = atof(strtok(buf+1, ":")) / 60.0;
				*t = (atof(strtok(NULL, ":")) -
					last_depth) / *r;
				return(1);
			}
			else *r = atof(buf+1) / 60.0;
			break;

		default:
			
			/* depth value input */

			if (strlen(buf)>0)
				d_target = atof(buf) / ucf;
			else d_target = last_depth;
			*r = 0.0;
	}

	for (*t = 0.0; *t == 0.0;) {
		move(STATS+3,15);
		clrtoeol();
		sprintf(buf,"Duration (default: %s): ",
			showtime(trans_time(last_depth, d_target)));
		strcpy(buf, prompt(buf));
		if (buf[0]=='q') return(0);
		if (strlen(buf)>0) *t = readtime(buf);
		else *t = trans_time(last_depth, d_target);
	}
	move(STATS+3,15);
	clrtoeol();
	*d = d_target;

	if (iterative && !running) {
		duration2 = *t - s;
		*t = s;
		running = ON;
	}
	return(1);
}

float time_to_surface(d, time_at_stop) 
float d, *time_at_stop;
{
	int i, j;
	float time=0, ctemp[NCOMP], stop, next, pp_inert_stop, pp_inert_next,
	      target, temp, dt, trans_time(), t_ascent, ceiling,
	      calc_ceiling();

	record_type vals;

	record_query(total_time, &vals);

	/* if not in decompression, simply return the ascent time to
	reach the surface */

	ceiling = calc_ceiling(vals.comp, model);
	if (ceiling == 0.0) return(trans_time(d, 0.0));

	/* zero out time to first stop as flag that it has not been
	determined yet */

	*time_at_stop=0.0;

	/* make a copy of the compartment N2 partial pressure values */

	for (i=0; i<ncomp; i++) ctemp[i] = vals.comp[i];
	
	/* set up first stop data. We must account for ascent time to the
	first stop, and acount for changes in N2 loading during the
	ascent.  */

	stop = ceil((double)(ceiling/5.0))*5.0;
	pp_inert_stop = tank.mix * (stop+ONE_ATM_FSW);
	time = trans_time(d, stop);
	update_tissues(ctemp, d, stop, time);


	/* for each stop, calculate the time to get to it from the
	previous stop. This involves calculating the time to decompress
	sufficiently to ascend to the next stop, and adding up each of
	those times. */

	for (next=stop-5.0; next>=0.0; next-=5.0) {

		/* calculate the ambient ppN2 at the target stop */

		pp_inert_next = tank.mix * (next+ONE_ATM_FSW);

		/* now spin through each compartment, finding the
		compartment that will take the longest to reach its
		maximum allowed ppN2 at the target stop depth. This is
		the decompression time necessary to advance to the next
		stop */

		for (i=0, temp=0; i<ncomp; i++) {

			/* determine the time to decompress compartment
			to target N2 pressure. Target pressure is the
			value at which the compartment yields a ceiling
			equal to the next stop. Time is then calculated
			using the exponential decay formula with the
			current stop as the ambient pressure */

			target = model[i].M0 + next * model[i].dM;
			if ((target < ctemp[i])&&(pp_inert_stop < ctemp[i])) {
				dt = model[i].offgas / LN_HALF *
					log((double)(1.0 -
					(target - ctemp[i]) /
					(pp_inert_stop - ctemp[i])));
				if (dt > temp) temp = dt;
			}
		}

		if (*time_at_stop == 0.0) *time_at_stop=time + temp;

		/* note: the variable 'temp' will contain the MAX() of
		all the decompression times for each compartment */

		/* adjust compartment values for next iteration/stop */

		update_tissues(ctemp, stop, stop, temp);
		t_ascent = trans_time(stop, next);
		update_tissues(ctemp, stop, next, t_ascent);

		/* accumulate total time for ascent */

		time += temp + t_ascent;

		/* move to the next stop */

		stop = next;

		/* Adjust the ambient ppN2 to reflect this new stop */

		pp_inert_stop = tank.mix * (stop+ONE_ATM_FSW);
	}
	return(time);
}

float ndl(d, comp, m, control, mix)
float d, comp[], mix;
model_type m[];
int *control;
{
	int i;

	float pp_inert, tmp, ndl_time;

	ndl_time = INF;
	*control = NOCONTROL;
	pp_inert = mix * (d + ONE_ATM_FSW);

	for (i=0; i<ncomp; i++) {

		if (comp[i] > m[i].M0) return(0.0);

		else if ((comp[i] < pp_inert) &&
			 (pp_inert > m[i].M0)) {

			/* compartment is ingassing, and the inter gas
			partial pressure is greater than the M0 for the
			compartment, both requirements for controlling
			the dive. Calculate the time to ingas to the M0
			at deph 'd', and adjust the candidate NDL and
			controlling compartment if shorter than the
			current projected NDL */

			tmp = m[i].offgas / LN_HALF *
				log((pp_inert - m[i].M0) /
				    (pp_inert - comp[i]));

			if (tmp < ndl_time) {
				ndl_time = tmp;
				*control = i;
			}
		}
	}
	return(ndl_time);
}

int buffer=7;
char *showtime(time)
float time;
{
	char *strcpy();

	static char buf[8][64];

	int t;

	t = time;
	buffer = (buffer + 1) % 8;

	if (t >= INF) return(strcpy(buf[buffer],"INF"));

	if (t >= 0) sprintf(buf[buffer], "%02d:%02d:%02d", t / 3600,
		(t % 3600) / 60, t % 60);
	else sprintf(buf[buffer], "-%02d:%02d:%02d", -t / 3600,
		(-t % 3600) / 60, -t % 60);
	return(buf[buffer]);
}

double deco_time(d, c, m)
float d, c;
model_type *m;
{
	double temp, pp_inert;

	pp_inert = tank.mix * (d + ONE_ATM_FSW);

	temp = (pp_inert - m->M0) / (pp_inert - c);
	return(m->offgas / LN_HALF / 60.0 * log(temp));
}

float loading(comp, M0)
float comp, M0;
{
	return((comp - one_atm_pp_inert) / (M0 - one_atm_pp_inert));
}

getaline(fp,buf)
FILE *fp;
char *buf;
{
	int c;
	int i=0;

	while (((c=getc(fp))!='\n')&&(c!=EOF)) buf[i++]=c;
	buf[i] = '\0';
	if (c==EOF) return(EOF);
	else return(i);
}

float trans_time(d1, d2)
float d1, d2;
{
	int i;
	float total, t, find_rate(), rate;
	double fabs();

	/* first handle the degenerate case of d1==d2 */

	if (d1 == d2) return(0.0);

	for (total=0.0; d1 != d2; ) {
		rate = find_rate(d1, d2, &t);
		total += t;
		d1 += rate * t;
		d1 = (d1 < fabs(rate) ? d2 : d1);
	}
	return(total);
}

float calc_ceiling(comp, m)
float comp[];
model_type m[];
{
	int i;
	float ceiling;

	for (i=0, ceiling=0.0; i<ncomp; i++)
		if (((comp[i] - m[i].M0) / m[i].dM) > ceiling)
			ceiling = (comp[i] - m[i].M0) / m[i].dM;
	return(ceiling);
}

record_query(t, rs)
float t;
record_type *rs;
{
	int i, j;
	float elapsed, calc_air();

	/* handle the degenerate cases first */

	if (num_records == 0) {
		rs->elapsed = 0;
		rs->depth = 0;
		rs->rate = 0;
		rs->duration = 0;
		for (i=0; i<ncomp; i++) rs->comp[i] = ONE_ATM_FSW_N2;
		rs->air.rmvs = rmvs;
		rs->air.tank = tank;
		rs->decoflag = FALSE;
		return;
	}
	if (t <= 0.0) {
		*rs = record[0];
		return(0);
	}

	/* find the appropriate segment from the depth record which
	includes t */

	for (	i=0;
		(i < num_records - 1) &&
		(t >= record[i].elapsed + record[i].duration);
		i++
	    );

	/* copy the record first, then adjust the parameters that are
	not fixed throughout the duration of the record */

	*rs = record[i];
	rs->elapsed = t;
	rs->depth = record[i].depth + record[i].rate * (t - record[i].elapsed);
	for (j=0; j<ncomp; j++) rs->comp[j] = record[i].comp[j];
	adjust_loading(rs->comp, model, record[i].depth,
		record[i].rate, t - record[i].elapsed,
		&record[i].air.tank);
	for (rs->decoflag=FALSE, j=0; j<ncomp; j++)
		if (rs->comp[j] > model[j].M0) rs->decoflag = TRUE;

	rs->air.tank = record[i].air.tank;
	if ((record[i].depth > 0.0) || (record[i].rate != 0.0))
		rs->air.tank.pres -= calc_air(record[i].depth,
					      record[i].rate,
					      t - record[i].elapsed,
					      record[i].air.rmvs) /
                                     record[i].air.tank.cap *
				     record[i].air.tank.wp;
	return(i);
}

void analysis()
{
	void clear_stats_region(), adjust_record();

	char *showtime(), c1, inbuf[64], tmpbuf[64];

	float start, t, inc, readtime(), d0, update_tissues(),
		elapsed, res=60.0, old_res, dur, temp, d_inc=10.0;

	record_type vals, vals2;

	int numchars=0, i, j, cursor;

	void display_air();	/* WC 5/21/98 */

	t = total_time;
	start = floor((t-3600.0<0.0?0.0:t-3600.0) / 60.0) * 60.0;
	inc = 60.0;

	for (;;) {
		if (mark != NOMARK)
			record_query(mark, &vals);
		else
			record_query(t, &vals);
		if (!numchars) {
			display_loading(vals.comp, vals.depth,
				vals.air.tank.mix);

			display_air(&vals);

			clear_stats_region();

			move(STATS+1,15);
			sprintf(tmpbuf, "Cursor increments: Time: %s ; Depth: %.1f %s", showtime(inc),
				d_inc * ucf, unitstr);
			addstr(tmpbuf);
			sprintf(tmpbuf,
				"t = %s ; Depth: %3.1f %s ; Ascent rate: %3.1f %s/min",
				showtime(t), vals.depth, unitstr,
				60.0 * -vals.rate * ucf, unitstr);
			move(STATS+2,15);
			addstr(tmpbuf);

			if ((temp=calc_ceiling(vals.comp, model)) > 0.0)
			{
				move(ncomp, 60);
				sprintf(tmpbuf, "Ceiling: %.1f", temp);
				addstr(tmpbuf);
			}

			if (t - start > res*60) start = t - res*60;
			if (t < start) start = t;
			if (inc < res) {
				old_res = res;
				res = inc;
				start =  t - ((t - start) / old_res) * res;
			}

			/* round the starting value for the profile
			display to the nearest multiple of the
			resolution. This fixes any problems with the
			cursor being inbetween points on the display */

			start = floor(start / res + 0.5) * res;
			update_dive_profile(start, t, res);
			if (mark == NOMARK) update_ndl_limits(t);
			else update_ndl_limits(mark);
		}
		sprintf(tmpbuf, "%Input: %-19s ; Resolution: %s",
                        (numchars?inbuf:""), showtime(res));
		move(STATS+3,15);
		clrtoeol();
		addstr(tmpbuf);

		refresh();
		
		/* get kbd input from user */

		switch (c1=getch()) {
			case ',':
			case '.':
				if (numchars > 0)
					inc = readtime(inbuf);
				t += (c1=='.' ? inc : -inc);
				numchars = 0;
				break;

			case 'g':
				if (numchars) t = readtime(inbuf);
				else t = ZBOUND(mark);
				start = (t < total_time ? t : total_time)
					- res * 30;
				start = ZBOUND(start);
				numchars = 0;
				break;

			case 'l':
				if (numchars > 0) {
					if (atoi(inbuf)>ncomp+1) {
						numchars = 0;
						break;
					}
					plot_tissue =
						atoi(inbuf);
					if (plot_tissue > ncomp)
						plot_tissue = ncomp;
					numchars = 0;
					graph_loading = ON;
				}
				else graph_loading =
					(graph_loading?OFF:ON);
				break;

			case 'r':
				if (numchars > 0) {
					old_res = res;
					res = readtime(inbuf);
					start =  t - ((t - start) /
						old_res) * res;
					numchars = 0;
				}
				break;

			case 'd':
				if (numchars > 0) {
					d_inc = atof(inbuf);
					numchars = 0;
				}
				break;

			case 'm':
				plot_mvalue = (plot_mvalue?OFF:ON);
				break;

			case 'M':
				mark = (mark == NOMARK ? t : NOMARK);
				break;

			case '>':
			case '<':
			case '^':
			case 'v':
				if (mark != NOMARK) adjust_record(&t, c1, inc,
					d_inc);
				else {
					clear_stats_region();
					move(STATS+1, 15);
					addstr("You must set a mark before edit operations");
					move(0,0);
					refresh();
					sleep(5L);
				}
				break;

			case 'q':
			case 'Q':
				mark = NOMARK;
				record_query(total_time, &vals);
				display_loading(vals.comp, vals.depth,
					vals.air.tank.mix);
				display_air(&vals);
				update_dive_profile(total_time - 3600.0,
					total_time, 60.0);
				update_ndl_limits(total_time);
				refresh();
				return;

			default:
				if (c1 == '\010') {
					if (numchars) numchars--;
				}
				else inbuf[numchars++] = c1;
				inbuf[numchars]='\0';
		}
		if (t < 0.0) t = 0.0;
		if (t > total_time) t = total_time;
	}
}

float readtime(buf)
char *buf;
{
	float t=0.0;

	int i, len;

	char *cptr, *strtok(), ibuf[16], *strcpy();

	double atof();

	/* we must make a copy of the input string, because the use of
	strtok() in this procedure can interfere with use of strtok() in
	the calling procedure if the string passed in is part of a
	buffer that is being parsed with strtok() */

	strcpy(ibuf, buf);

	if (strchr(ibuf,':') == NULL) {

		/* string is in simple time format. This format can only
		be used to specify an integral number of hours, minutes,
		or seconds; no combination of units is allowed. For
		combination specifications (like 2 hours, 3 minutes, and
		30 seconds) must be specified in hh:mm:ss format. */

		len = strlen(ibuf);
		if (ibuf[len-1] == 'h') {
			ibuf[len-1] = '\0';
			return(atof(ibuf)*3600.0);
		}
		else if (ibuf[len-1] == 'm') {
			ibuf[len-1] = '\0';
			return(atof(ibuf)*60.0);
		}
		else if (ibuf[len-1] == 's') {
			ibuf[len-1] = '\0';
			return(atof(ibuf));
			
		}
		else return(atof(ibuf)*60.0);
	}
	else t = atof(strtok(ibuf,":"));

	while ((cptr = strtok((char *)NULL,":")) != NULL)
	{
		t = t * 60.0 + atof(cptr);
	}

	return(t);
}

float calc_air(d, r, t, rmvs)
float d, r, t, rmvs;
{
	return(rmvs * (r * t * t / 66.0 + d * t / 33.0 + t));
}

char *get_menu_choice(msg, menu)
char *msg;
menu_type *menu;
{
	char *tmp, *buf, chr;	/* removed declaration of malloc().  WC 5/21/98 */

	int i, n_items, maxlen, cmds_per_line;

	static int item_ptr=0;

	menu_type *mptr, *selection;

	static menu_type *last_menu=NULL;

	void clear_stats_region();

	clear_stats_region();

	move(STATS, 15);
	addstr(msg);

	if (last_menu != menu) item_ptr=0;
	last_menu = menu;

	/* determine the number of items in the menu, and the widest
	   entry for display calculations */

	for (mptr=menu, n_items=0, maxlen=0; mptr!=NULL; n_items++) {
		maxlen = MAX(maxlen, strlen(mptr->label));
		mptr = mptr->next;
	}	

	maxlen += 3;
	cmds_per_line = (width - 15) / maxlen;

	for (;;) {

		/* Display the menu */

		for (mptr=menu, i=0; mptr!=NULL; i++) {
			move(STATS+1 + i / cmds_per_line,
				15 + (i % cmds_per_line) * maxlen);
			if (item_ptr == i) {
				selection = mptr;
				addstr("->");
			}
			else addstr("  ");
			addstr(mptr->label);
			mptr = mptr->next;
		}
		move(0,0);
		refresh();

		/* get input from the user and process */

		switch(chr=getch()) {
			case '.':
				item_ptr = (item_ptr + 1) % n_items;
				break;
			
			case ',':
				item_ptr = (item_ptr + n_items - 1) % n_items;
				break;

			case '\n':
				strcpy((buf=malloc(80)), selection->label);
				if (selection->submenu != NULL) {
					tmp = get_menu_choice(msg, selection->submenu);
					strcat(buf, "/");
					strcat(buf, tmp);
					free(tmp);
				}
				return(buf);
		}
	}
}

void display_air(rp)
record_type *rp;
{
	int i;
	char buf[80], *showtime();

	for (i=0; i<6; i++) {
		move(LGRAPH+i+1, 60);
		clrtoeol();
	}

	sprintf(buf, "Tank type: %s", rp->air.tank.label);
	move(LGRAPH+1, 60);
	addstr(buf);

	sprintf(buf, "Press: %.1f %s",
		rp->air.tank.pres /
		(rp->air.tank.type==ENGLISH?1.0:PSI_PER_BAR),
		(rp->air.tank.type==ENGLISH?"psi":"bar"));
	move(LGRAPH+2, 60);
	addstr(buf);

	sprintf(buf, "At %.1f %s/min,", rp->air.rmvs * 60.0 *
		(rp->air.tank.type==ENGLISH?1.0:L_PER_CUFT),
		(rp->air.tank.type==ENGLISH?"cuft":"liter"));
	move(LGRAPH+3, 60);
	addstr(buf);

	sprintf(buf, "%s remaining",
		showtime(ceil(rp->air.tank.pres / rp->air.tank.wp *
			      rp->air.tank.cap / rp->air.rmvs *
			      ONE_ATM_FSW / (rp->depth + ONE_ATM_FSW))));
	move(LGRAPH+4, 60);
	addstr(buf);
}

void adjust_record(t, type, t_inc, d_inc)
float *t, t_inc, d_inc;
char type;
{
	record_type vals;
	int rec, mark_rec, prev_dc, post_dc, i, j;
	float calc_air(), mark_offset;
	double floor();

	/* find out which record entry contains the specified 't' value */

	mark_rec = record_query(mark, &vals);
	mark_offset = mark - record[mark_rec].elapsed;
	rec = record_query(*t, &vals);

	/* now scan backwards through the record for the clostest
	previous constant depth segment -- this is the segment over
	which editing operations will be applied. */

	for (; (record[rec].rate != 0.0) && (rec > 0); rec--);
	if (record[rec].rate != 0.0) return;

	prev_dc = rec;

	switch (type) {

		case '<':
			t_inc = -t_inc;
		case '>':
			record[rec].duration += t_inc;

			/* if the cursor is after the end of the
			segment being adjusted, it needs adjustment as
			well to keep it in the same place */

			if (*t >= record[rec+1].elapsed) *t += t_inc;
			break;

		case '^':
			d_inc = -d_inc;
		case 'v':

			/* the ascent records before and after the
			specified constant depth segment need to be
			updated so that their durations and initial
			depths are correct. Scan backward first to find
			the previous ascent record, change it, and scan
			forward to find the following ascent record */

			/* find the previous depth change record */

			for (prev_dc=rec-1;
			     (prev_dc>0) && (record[prev_dc].rate==0.0);
			     prev_dc--);

			/* find the following depth change record, if
			any */

			for (post_dc=rec+1; (post_dc<num_records) &&
				      (record[post_dc].rate==0.0); i++);
			
			/* the depth change is allowed as long as the
			change will not reduce the duration of either
			depth change record to less than zero */

			if ((record[prev_dc].duration + d_inc /
				record[prev_dc].rate < 0.0) ||
				((record[post_dc].duration - d_inc /
				record[post_dc].rate < 0.0) &&
				post_dc < num_records)) return;

			/* adjust the depth entry in the record */

			record[rec].depth += d_inc;

			/* adjust the durations of the previous and
			following depth change records, if any, to
			reflect the new depth */

			record[prev_dc].duration += d_inc /
				record[prev_dc].rate;
			if (*t >= record[prev_dc+1].elapsed)
				*t += d_inc / record[prev_dc].rate;

			if (post_dc < num_records) {
				record[post_dc].duration -= d_inc /
					record[post_dc].rate;
				if (*t >= record[post_dc+1].elapsed)
					*t -= d_inc / record[post_dc].rate;
			}
	}

	/* update the rest of the record to make it consistent with the
	changes */

	for (i=prev_dc+1; i<num_records; i++) {
		record[i].elapsed = record[i-1].elapsed +
			record[i-1].duration;
		record[i].depth = floor(100.0 * (record[i-1].depth +
			record[i-1].rate * record[i-1].duration)) / 100.0;
		for (j=0; j<ncomp; j++)
			record[i].comp[j] = record[i-1].comp[j];
			adjust_loading(
				record[i].comp,
				model,
				record[i-1].depth,
				record[i-1].rate,
				record[i-1].duration,
				&record[i-1].air.tank);
		
		if (!strcmp(record[i].air.tank.label,
			    record[i-1].air.tank.label)) {

			record[i].air.tank.pres = record[i-1].air.tank.pres;

			if (	(record[i-1].depth > 0.0) ||
				(record[i-1].rate != 0.0)
			) record[i].air.tank.pres -=
				calc_air(record[i-1].depth,
				         record[i-1].rate,
					 record[i-1].duration,
					 record[i-1].air.rmvs) /
				record[i-1].air.tank.cap *
				record[i-1].air.tank.wp;
		}
	}
	total_time = record[num_records-1].elapsed +
		record[num_records-1].duration;
	mark = record[mark_rec].elapsed + mark_offset;
}

FILE *open_file(msg, mode)
char *msg, *mode; 
{
	FILE *fp, *fopen();
	char buf[64], *bptr, *prompt();

	for (fp=NULL; fp == NULL; ) {
		bptr = prompt(msg);
		if (!strlen(bptr)) break;
		if (!(fp=fopen(bptr,mode))) {
			move(STATS+2,15);
			clrtoeol();
			sprintf(buf,"Couldn't open file \"%s\"... Try again",
				buf);
			addstr(buf);
		}
	}
	return(fp);
}

menu_index(menu, choice)
menu_type *menu;
char *choice;
{
	int i=0;

	menu_type *mptr;

	for (	i = 0, mptr = menu;
		(mptr!=NULL) && strcmp(mptr->label, choice);
		i++, mptr = mptr->next
	);

	if (mptr != NULL) return(i);
	else return(-1);
}

void air_ops(choice)
char *choice;
{
	FILE *fp;
	tank_type *ptr;
	menu_type *mptr;
	char buf[64], *prompt(), *get_menu_choice(), label[16];
	float cap, wp, mix;
	void display_air();

	if (!strcmp(choice, "Select")) {
		choice = get_menu_choice("Select:", tank_menu);
		for (
			ptr = tank_models;
			(ptr != NULL) && strcmp(ptr->label, choice);
			ptr = ptr->next
		);
		if (ptr != NULL) {
			tank = *ptr;
			current_tank_model = ptr;
		}
		new_tank = TRUE;
	}
	else if (!strcmp(choice, "Reset")) {
		tank = *current_tank_model;
		new_tank = TRUE;
	}
	else if (!strcmp(choice, "RMVS")) {
		rmvs = atof(prompt("Enter breathing rate: ")) / 60.0;
		if (tank.type == METRIC) rmvs /= L_PER_CUFT;
		if (num_records == 0) record[0].air.rmvs = rmvs;
	}
	else if (!strcmp(choice, "Load")) {
		fp = open_file("Tank model input file: ", "r");
		while (getaline(fp, buf) != EOF) {
			strcpy(label, strtok(buf, ":"));
			cap = atof(strtok(NULL, ":"));
			wp = atof(strtok(NULL, ":"));
			mix = atof(strtok(NULL, ":"));
			for (
				ptr = tank_models;
				(ptr != NULL) && strcmp(label, ptr->label);
				ptr = ptr->next
			);

			if (ptr == NULL) {

				/* allocate a new tank structure for the
				tank spec being loaded, and link it into
				the list */

				ptr = last_tank_entry;
				ptr->next = malloc(sizeof(tank_type));
				ptr = ptr->next;
				ptr->next = NULL;
				last_tank_entry = ptr;
				strcpy((ptr->label=malloc(16)), label);

				/* create menu entry for the tank */

				mptr = last_tank_menu_entry;
				mptr->next = malloc(sizeof(menu_type));
				mptr = mptr->next;
				mptr->next = NULL;
				last_tank_menu_entry = mptr;
				mptr = ptr->label;
			}
			ptr->cap = cap;
			ptr->wp = ptr->pres =  wp;
			ptr->mix = mix;

		}
		fclose(fp);
	}
	else if (!strcmp(choice, "Dump")) {
		fp = open_file("Tank model output file: ", "w");
		for (ptr = tank_models; ptr != NULL; ptr = ptr->next)
			fprintf(
				fp,
				"%s:%.1f:%.1f:%.2f\n",
				ptr->label,
				ptr->cap,
				ptr->wp,
				ptr->mix
			);
		fclose(fp);
	}
	else if (!strcmp(choice, "Build")) {
		clear_stats_region();

		/* allocate space for new tank structure, and
		link into list */

		ptr = last_tank_entry;
		ptr->next = malloc(sizeof(tank_type));
		ptr = ptr->next;
		ptr->next = NULL;
		last_tank_entry = ptr;

		/* allocate space for a new menu entry in the
		tank selection menu, and link it on to the end
		of the list */

		mptr = last_tank_menu_entry;
		mptr->next = malloc(sizeof(menu_type));
		mptr = mptr->next;
		mptr->next = NULL;
		last_tank_menu_entry = mptr;

		/* prompt for various parameters, and fill in
		the model */

		strcpy((ptr->label=malloc(16)),
			prompt("Tank label: "));
		mptr->label = ptr->label;

		strcpy(buf, get_menu_choice("Select a tank type:",
			type_menu));
		if (!strcmp(buf, "English")) ptr->type = ENGLISH;
		else ptr->type = METRIC;

		if (ptr->type == ENGLISH)
			ptr->cap = atof(prompt("Capacity: "));
		else ptr->cap = atof(prompt("Volume: ")) / L_PER_CUFT;

		ptr->wp = atof(prompt("Working pressure: "));
		if (ptr->type == METRIC) {
			ptr->wp *= PSI_PER_BAR;
			ptr->cap *= ptr->wp / ONE_ATM_PSI;
		}
		ptr->pres = ptr->wp;

		ptr->mix = atof(prompt("Inert gas %: ")) / 100.0;
	}
	else if (!strcmp(choice, "Pres")) {
		tank.pres = atof(prompt("Tank pressure: "));
		new_tank = TRUE;
	}
}

char *prompt(msg)
char *msg;
{
	static char buf[128];

	echo();
	move(STATS+2, 15);
	clrtoeol();
	addstr(msg);
	refresh();
	getstr(buf);
	noecho();
	return(buf);
}

void depth_range(t, d_start, d_end)
float t, *d_start, *d_end;
{
	static float cur_start=0.0;
	float cur_end;
	record_type vals;

	cur_end = cur_start + (lines - ncomp - 6) * di;
	record_query(t, &vals);

	if ((vals.depth >= cur_start) && (vals.depth <= cur_end)) {
		*d_start = cur_start;
		*d_end = cur_end;
		return;
	}

	if (vals.depth < cur_start) {
		cur_start = floor(vals.depth / di) * di;
		cur_end = cur_start + (lines - ncomp - 6) * di;
	}
	else if (vals.depth > cur_end) {
		cur_end = ceil(vals.depth / di) * di;
		cur_start = cur_end - (lines - ncomp - 6) * di;
	}
	*d_start = cur_start;
	*d_end = cur_end;
}

void display_depths(start, end)
float start, end;
{
	float f;
	static float last_start=0.0, last_end=0.0;
	int y, i;
	char buf[16];

	if ((start == last_start) && (end == last_end)) return;
	for (f=start; f <= end; f+=di) {
		y = ncomp + 5 + (f - start) / di;
		/* move(y, 4);
		addstr("      "); */
		move(y, 4);
		sprintf(buf,"%5.1f|",f*ucf);
		addstr(buf);
	}
	last_start = start;
	last_end = end;
}
